<?php

namespace FiloBlu\Esb\Model;

use DateTime;
use Exception;
use FiloBlu\Esb\Api\ConsumerConfigurationRepositoryInterface;
use FiloBlu\Esb\Api\ConsumerResolverInterface;
use FiloBlu\Esb\Api\Data\ConsumerConfigurationInterface;
use FiloBlu\Esb\Api\Data\QueueItemInterface;
use FiloBlu\Esb\Api\Data\QueueMetadataInterface;
use FiloBlu\Esb\Api\Data\QueueMetadataInterfaceFactory;
use FiloBlu\Esb\Api\Data\StatusInterface;
use FiloBlu\Esb\Api\Data\StatusInterfaceFactory;
use FiloBlu\Esb\Api\DispatcherInterface;
use FiloBlu\Esb\Api\PublisherResolverInterface;
use FiloBlu\Esb\Api\QueueItemRepositoryInterface;
use FiloBlu\Esb\Core\Exception\NonRecoverableException;
use Magento\Framework\Api\SearchCriteriaBuilder;
use Magento\Framework\Api\SearchCriteriaBuilderFactory;
use Psr\Log\LoggerInterface;
use Throwable;

/**
 * Class Dispatcher
 * @package FiloBlu\Esb\Model
 */
class Dispatcher implements DispatcherInterface
{
    /**
     * @var QueueItemRepositoryInterface
     */
    protected $queueItemRepository;
    /**
     * @var SearchCriteriaBuilderFactory
     */
    protected $searchCriteriaBuilderFactory;
    /**
     * @var LoggerInterface
     */
    protected $logger;
    /**
     * @var ConsumerResolverInterface
     */
    private $consumerResolver;

    /**
     * @var PublisherResolverInterface
     */
    private $publisherResolver;
    /**
     * @var ConsumerConfigurationRepositoryInterface
     */
    private $consumerConfigurationRepository;
    /**
     * @var QueueMetadataInterfaceFactory
     */
    private $queueMetadataFactory;
    /**
     * @var StatusInterfaceFactory
     */
    private $statusFactory;

    /**
     * Dispatcher constructor.
     * @param LoggerInterface $logger
     * @param ConsumerResolverInterface $consumerResolver
     * @param PublisherResolverInterface $publisherResolver
     * @param QueueItemRepositoryInterface $queueItemRepository
     * @param SearchCriteriaBuilderFactory $searchCriteriaBuilderFactory
     * @param ConsumerConfigurationRepositoryInterface $consumerConfigurationRepository
     * @param QueueMetadataInterfaceFactory $queueMetadataFactory
     * @param StatusInterfaceFactory $statusFactory
     */
    public function __construct(
        LoggerInterface $logger,
        ConsumerResolverInterface $consumerResolver,
        PublisherResolverInterface $publisherResolver,
        QueueItemRepositoryInterface $queueItemRepository,
        SearchCriteriaBuilderFactory $searchCriteriaBuilderFactory,
        ConsumerConfigurationRepositoryInterface $consumerConfigurationRepository,
        QueueMetadataInterfaceFactory $queueMetadataFactory,
        StatusInterfaceFactory $statusFactory
    ) {
        $this->searchCriteriaBuilderFactory = $searchCriteriaBuilderFactory;
        $this->queueItemRepository = $queueItemRepository;
        $this->logger = $logger;
        $this->consumerResolver = $consumerResolver;
        $this->publisherResolver = $publisherResolver;
        $this->consumerConfigurationRepository = $consumerConfigurationRepository;
        $this->queueMetadataFactory = $queueMetadataFactory;
        $this->statusFactory = $statusFactory;
    }

    /**
     *
     * @throws Exception
     */
    public function dispatch()
    {
        /** @var SearchCriteriaBuilder $builder */
        $builder = $this->searchCriteriaBuilderFactory->create();
        $builder->addFilter(QueueItemInterface::STATUS, StatusInterface::PENDING);
        $builder->addFilter(
            QueueItemInterface::EXECUTABLE_AFTER,
            (new DateTime())->format(QueueItemRepositoryInterface::DATE_FORMAT),
            'lteq'
        );
        $builder->setSortOrders([
            QueueItemInterface::EXECUTABLE_AFTER => 'asc',
            QueueItemInterface::PUBLISHED_AT => 'asc',
            QueueItemInterface::PRIORITY => 'asc'
        ]);

        $result = $this->queueItemRepository->getList($builder->create());

        $publisher = $this->publisherResolver->getDefault();

        /** @var QueueItemInterface $item */
        foreach ($result->getItems() as $item) {

            // Dispatch to all consumers
            foreach ($this->getConsumersToProcess($item) as $consumerConfiguration) {
                $status = $item->getStatus();

                try {
                    $this->setProcessing($item);
                    $consumer = $consumerConfiguration->getConsumer();
                    $message = $consumer->consume(clone $item->getMessage());

                    $outputEvent = $consumerConfiguration->getOutputEvent();

                    // TODO : this must be done into a validator
                    if ($outputEvent && $outputEvent !== $item->getEvent()) {
                        /** @var QueueMetadataInterface $metadata */
                        $metadata = $this->queueMetadataFactory->create();
                        $metadata->setFrom($item->getFrom());
                        // TODO : recalculate consumer binded to event
                        $metadata->setTo($item->getTo());
                        $publisher->publish($message, $consumerConfiguration->getOutputEvent(), $metadata);
                    }

                    $status = $consumer->getStatus();

                    switch ($status->getCode()) {
                        case StatusInterface::SKIPPED:
                        case StatusInterface::SUCCESS:
                            $item->setStatus($status);
                            break;
                        default:
                            $status = $this->handleError($item, $consumerConfiguration);
                    }
                } catch (NonRecoverableException $nonRecoverableException) {
                    $status->setOutputData($nonRecoverableException->getMessage())->setCode(StatusInterface::ERROR);
                } catch (Exception $exception) {
                    $status = $this->handleFailure($item, $consumerConfiguration, $exception);
                } catch (Throwable $throwable) {
                    $status = $this->handleFailure($item, $consumerConfiguration, $throwable);
                } finally {
                    $item->setFinishedAt(new DateTime());
                    $item->setStatus($status);
                    $this->queueItemRepository->save($item);
                }
            }
        }
    }

    /**
     * @param QueueItemInterface $queueItem
     * @return ConsumerConfigurationInterface[]
     */
    protected function getConsumersToProcess(QueueItemInterface $queueItem): array
    {
        /** @var SearchCriteriaBuilder $searchCriteriaBuilder */
        $searchCriteriaBuilder = $this->searchCriteriaBuilderFactory->create();
        $searchCriteria = $searchCriteriaBuilder
            ->addFilter(ConsumerConfigurationInterface::ENABLE, ConsumerConfigurationInterface::STATUS_ENABLED)
            ->addFilter(ConsumerConfigurationInterface::BINDING_EVENT, $queueItem->getEvent())
            ->addFilter(ConsumerConfigurationInterface::UUID, $queueItem->getTo())
            ->create();

        $result = $this->consumerConfigurationRepository->getList($searchCriteria);
        return $result->getItems();
    }

    /**
     * @param QueueItemInterface $queueItem
     * @throws Exception
     */
    protected function setProcessing(QueueItemInterface $queueItem)
    {
        $queueItem->getStatus()->setCode(StatusInterface::PROCESSING);
        $queueItem->setExecutedAt(new DateTime());
        $this->queueItemRepository->save($queueItem);
    }

    /**
     * @param QueueItemInterface $item
     * @param ConsumerConfigurationInterface $consumerConfiguration
     * @return StatusInterface
     */
    protected function handleError(QueueItemInterface $item, ConsumerConfigurationInterface $consumerConfiguration)
    {
        $id = $item->getId();

        if ($consumerConfiguration->getRetryStrategy()->reschedule($item)) {
            $message = "Rescheduled item with id {$id} because an error occurred";
        } else {
            $message = "Item with {$id} will not be scheduled anymore";
        }

        $this->logger->error($message);

        return $item->getStatus()->setCode(StatusInterface::ERROR);
    }

    /**
     * @param QueueItemInterface $item
     * @param ConsumerConfigurationInterface $consumerConfiguration
     * @param Throwable $exception
     * @return StatusInterface
     */
    protected function handleFailure(QueueItemInterface $item, ConsumerConfigurationInterface $consumerConfiguration, Throwable $exception): StatusInterface
    {
        $id = $item->getId();

        $message = $exception->getMessage();

        if ($consumerConfiguration->getRetryStrategy()->reschedule($item)) {
            $message = "Rescheduled item with id {$id} because exception : {$message}";
        } else {
            $message = "Item with {$id} will not be scheduled anymore : {$message}";
        }

        $this->logger->error($message, ['exception' => $exception]);

        return $item->getStatus()->setOutputData($exception->getMessage())->setCode(StatusInterface::ERROR);
    }
}
